#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Mapper classes."""
from typing import Dict, List, Tuple

from esdl.esdl import Joint as esdl_junction

from omotes_simulator_core.adapter.transforms.esdl_asset_mapper import EsdlAssetMapper
from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract

from omotes_simulator_core.entities.assets.junction import Junction
from omotes_simulator_core.entities.esdl_object import EsdlObject
from omotes_simulator_core.entities.heat_network import HeatNetwork
from omotes_simulator_core.entities.network_controller import NetworkController
from omotes_simulator_core.simulation.mappers.mappers import EsdlMapperAbstract
from omotes_simulator_core.solver.network.network import Network
from omotes_simulator_core.adapter.transforms.esdl_asset_mapper import (
    EsdlAssetControllerProducerMapper,
)
from omotes_simulator_core.adapter.transforms.esdl_asset_mapper import (
    EsdlAssetControllerConsumerMapper,
)
from omotes_simulator_core.adapter.transforms.esdl_asset_mapper import (
    EsdlAssetControllerStorageMapper,
)


def replace_joint_in_connected_assets(
    connected_py_assets: List[Tuple[str, str]],
    py_joint_dict: Dict[str, List[Tuple[str, str]]],
    py_asset_id: str,
    iteration_limit: int = 10,
) -> List[Tuple[str, str]]:
    """Replace joint with assets connected to the elements.

    Replace items in the connected_py_assets list that are connected to a Joint
    with the items that are connected to the Joint, except for the current asset.

    :param connected_py_assets: List of connected assets
    :param py_joint_dict: Dictionary with joint id as key and list of tuples with
    connected asset id and port as value.
    :param py_asset_id: Id of the asset all the connected assets are connected to
    :param iteration_limit: Limit for the number of iterations

    """
    replace_bool = True
    while replace_bool and iteration_limit > 0:
        replace_bool = False
        iteration_limit -= 1
        for index, connected_py_asset in enumerate(connected_py_assets):
            connected_py_asset_id, _ = connected_py_asset
            if connected_py_asset_id in py_joint_dict:
                # remove the joint from the list and add the connected assets
                connected_py_assets.pop(index)
                # get the connected assets
                additional_assets = py_joint_dict[connected_py_asset_id].copy()
                # remove the current asset from the list
                for index, additional_asset_properties in enumerate(additional_assets):
                    additional_asset_id, _ = additional_asset_properties
                    if additional_asset_id == py_asset_id:
                        additional_assets.pop(index)
                        break
                # add the connected assets to the list
                connected_py_assets.extend(additional_assets)
                replace_bool = True
                break
    if iteration_limit == 0:
        raise RuntimeError("Error in replacing joint in connected assets.")
    else:
        return connected_py_assets


class EsdlEnergySystemMapper(EsdlMapperAbstract):
    """Creates a HeatNetwork entity object based on a PyESDL EnergySystem object."""

    def __init__(self, esdl_object: EsdlObject):
        """Constructor for esdl to heat network mapper.

        :param esdl_object: Esdl object to be converted to a Heatnetwork
        """
        self.esdl_object = esdl_object

    def to_esdl(self, entity: HeatNetwork) -> EsdlObject:
        """Method to convert a HeatNetwork object back to an esdlobject.

        For now this method is not implemented.
        :param HeatNetwork entity: HeatNetwork object to be converted to esdl object
        :return: EsdlObject, which is the converted HeatNetwork object.
        """
        raise NotImplementedError("EsdlEnergySystemMapper.to_esdl()")

    def to_entity(self, network: Network) -> Tuple[List[AssetAbstract], List[Junction]]:
        """Method to convert esdl to Heatnetwork object.

        This method first converts all assets into a list of assets.
        Next to this a list of Junctions is created. This is then used
        to create the Heatnetwork object.
        :param Network network: network to add the components to.
        :return: (List[AssetAbstract], List[Junction]), tuple of list of assets and junctions.
        """
        # TODO: This method requires a clean-up!
        py_assets_list = self._convert_assets(network)
        py_junction_list = self._create_junctions(network, py_assets_list)

        return py_assets_list, py_junction_list

    def _create_junctions(
        self,
        network: Network,
        py_assets_list: list[AssetAbstract],
    ) -> List[Junction]:
        """Method to create junctions and connect the assets with them.

        :param network: network to add the junctions to.
        :param py_assets_list: list of assets to connect to the junctions.
        :param py_joint_dict: dictionary with all jints in the esdl.

        :return: List of junctions that are created and connected to the assets.
        """
        py_joint_dict = self._get_junction()
        py_junction_list = []
        # loop over assets and create junctions and connect them
        for py_asset in py_assets_list:
            for con_point in range(0, py_asset.number_of_con_points):
                if not py_asset.solver_asset.is_connected(con_point):
                    connected_py_assets = self.esdl_object.get_connected_assets(
                        py_asset.asset_id, py_asset.connected_ports[con_point]
                    )
                    # Replace items in the connected_py_assets list that are connected to a Joint
                    # with the items that are connected to the Joint, except for the current asset.
                    connected_py_assets = replace_joint_in_connected_assets(
                        connected_py_assets, py_joint_dict, py_asset.asset_id
                    )
                    for connected_py_asset, connected_py_port in connected_py_assets:
                        index = [py_asset_temp.asset_id for py_asset_temp in py_assets_list].index(
                            connected_py_asset
                        )
                        con_point_2 = py_assets_list[index].connected_ports.index(connected_py_port)

                        node_id = network.connect_assets(
                            asset1_id=py_asset.solver_asset.name,
                            connection_point_1=con_point,
                            asset2_id=py_assets_list[index].solver_asset.name,
                            connection_point_2=con_point_2,
                        )
                        py_junction_list.append(
                            Junction(network.get_node(node_id), name=str(node_id))
                        )
        return py_junction_list

    def _convert_assets(self, network: Network) -> List[AssetAbstract]:
        """Method to convert all assets from the esdl to a list of pyassets.

        This method loops over all assets in the esdl and converts them to pyassets.

        :param Network network: network to add the components to.
        :return: List of pyassets.
        """
        py_assets_list = []
        for esdl_asset in self.esdl_object.get_all_assets_of_type("asset"):
            # Esdl Junctions need to be skipped in this method, they are added in another method.
            if isinstance(esdl_asset.esdl_asset, esdl_junction):
                continue
            py_assets_list.append(EsdlAssetMapper.to_entity(esdl_asset))
            py_assets_list[-1].add_physical_data(esdl_asset=esdl_asset)
            network.add_existing_asset(py_assets_list[-1].solver_asset)
        return py_assets_list

    def _get_junction(self) -> dict[str, list[tuple[str, str]]]:
        """Method to create an overview of all assets connected to a joint in the esdl.

        This method creates a dictionary with the joint id as key.
        The value is a list of all connected assets and the id of the port it is connected to.

        :return: dict[Any, list[list[tuple[str, str]]], which is the dictionary with the connected
        assets.
        """
        py_joint_dict = {}
        for esdl_joint in self.esdl_object.get_all_assets_of_type("joint"):
            temp_list = [
                self.esdl_object.get_connected_assets(
                    asset_id=esdl_joint.esdl_asset.id, port_id=port
                )
                for port in esdl_joint.get_port_ids()
            ]
            py_joint_dict[esdl_joint.get_id()] = [item for sublist in temp_list for item in sublist]
        return py_joint_dict


class EsdlControllerMapper(EsdlMapperAbstract):
    """Creates a NetworkController entity object based on a PyESDL EnergySystem object."""

    def __init__(self) -> None:
        """Constructor for esdl to heat network mapper."""

    def to_esdl(self, entity: NetworkController) -> EsdlObject:
        """Method to convert a NetworkController object to an esdlobject.

        For now this method is not implemented.
        :param NetworkController entity: NetworkController object to be converted to esdl object
        :return: EsdlObject, which is the converted NetworkController object.
        """
        raise NotImplementedError("EsdlControllerMapper.to_esdl()")

    def to_entity(self, esdl_object: EsdlObject) -> NetworkController:
        """Method to convert esdl to NetworkController object.

        This method first converts all assets into a list of assets.
        Next to this a list of Junctions is created. This is then used
        to create the NetworkController object.
        :param EsdlObject esdl_object: esdl object to convert to NetworkController object.

        :return: NetworkController, which is the converted EsdlObject object.
        """
        consumers = [
            EsdlAssetControllerConsumerMapper().to_entity(esdl_asset=esdl_asset)
            for esdl_asset in esdl_object.get_all_assets_of_type("consumer")
        ]
        producers = [
            EsdlAssetControllerProducerMapper().to_entity(esdl_asset=esdl_asset)
            for esdl_asset in esdl_object.get_all_assets_of_type("producer")
        ]
        storages = [
            EsdlAssetControllerStorageMapper().to_entity(esdl_asset=esdl_asset)
            for esdl_asset in esdl_object.get_all_assets_of_type("storage")
        ]
        return NetworkController(producers, consumers, storages)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Esdl asset wrapper class."""

import logging
from typing import List, Tuple

from esdl.esdl_handler import EnergySystemHandler

from omotes_simulator_core.adapter.transforms.string_to_esdl import StringEsdlAssetMapper
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject

logger = logging.getLogger(__name__)


class EsdlObject:
    """EsdlObject class is a wrapper around PyEsdl."""

    energy_system_handler: EnergySystemHandler

    def __init__(self, esdl_energysystem_handler: EnergySystemHandler) -> None:
        """
        Constructor for EsdlObject.

        :param esdl_energysystem: PyEsdl EnergySystem object
        """
        self.energy_system_handler = esdl_energysystem_handler

    def __repr__(self) -> str:
        """Returns a string describing the esdl file object."""
        return str(self.energy_system_handler)

    def get_all_assets_of_type(self, esdl_asset_type: str) -> list[EsdlAssetObject]:
        """
        Returns a list of all the esdl assets of the specified type in the esdl file.

        If the type is not found an empty list is returned.
        :param esdl_asset_type: str of the asset type assets need to be gathered.
        """
        output_list = []
        for asset_type in StringEsdlAssetMapper().to_esdl(esdl_asset_type):
            output_list += [
                EsdlAssetObject(asset)
                for asset in self.energy_system_handler.get_all_instances_of_type(asset_type)
            ]
        return output_list

    def get_connected_assets(self, asset_id: str, port_id: str) -> List[Tuple[str, str]]:
        """Method to get the id's of connected assets from the esdl.

        This returns a list of a tuple with the id of the connected asset and the id of the port
        to which the original asset is connected.

        :param str asset_id: id of the asset for which to return the connected assets.
        :param str port_id: id of the port for which to return the connected assets.
        :return: List of tuple with the id of the connected assets and the connected port ids.
        """
        esdl_asset = self.energy_system_handler.get_by_id(asset_id)

        connected_port_ids = []
        for esdl_port in esdl_asset.port:
            if esdl_port.id == port_id:
                connected_port_ids = esdl_port.connectedTo
                break
        if not connected_port_ids:
            raise ValueError(f"No connected assets found for asset: {asset_id} and port: {port_id}")
        connected_assets = [(port.energyasset.id, port.id) for port in connected_port_ids]
        return connected_assets



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module for solving the network class."""
from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.matrix.matrix import Matrix
from omotes_simulator_core.solver.network.network import Network


class Solver:
    """Class to solve the network."""

    _iteration_limit: int = 100
    """The maximum number of iterations for the solver."""

    def __init__(self, network: Network):
        """Constructor of the solver class.

        Initializes the class properties and sets the unknowns of the matrix.

        :param Network network: The network to be solved.
        """
        self.matrix = Matrix()
        self.network = network
        self.set_unknowns_matrix()

    def set_unknowns_matrix(self) -> None:
        """Sets the unknowns of the matrix."""
        for asset in self.network.assets:
            self.network.get_asset(asset).set_matrix_index(
                self.matrix.add_unknowns(self.network.get_asset(asset).number_of_unknowns)
            )
        for node in self.network.nodes:
            self.network.get_node(node).set_matrix_index(
                self.matrix.add_unknowns(self.network.get_node(node).number_of_unknowns)
            )

    def get_equations(self) -> list[EquationObject]:
        """Method to get the equations of the network.

        :return: list[EquationObject] equations: List of equations of the network.
        """
        equations: list[EquationObject] = []
        for asset in self.network.assets:
            equations = equations + self.network.assets[asset].get_equations()
        for node in self.network.nodes:
            equations = equations + self.network.nodes[node].get_equations()
        return equations

    def solve(self) -> None:
        """Method to solve the network."""
        iteration = 0
        self.matrix.reset_solution()
        while not self.matrix.is_converged():
            iteration += 1
            equations = self.get_equations()
            self.matrix.solve(equations, dump=False)
            self.results_to_assets()
            if iteration > self._iteration_limit:
                print("No converged solution reached")
                break

    def get_results(self) -> None:
        """Method to get the results of the network."""

    def results_to_assets(self) -> None:
        """Method to transfer the results to the assets from the matrix."""
        self.network.set_result_asset(self.matrix.sol_new)
        self.network.set_result_node(self.matrix.sol_new)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""NetworkController entity."""
import datetime
import logging
from typing import List

from omotes_simulator_core.entities.network_controller_abstract import NetworkControllerAbstract
from omotes_simulator_core.entities.assets.asset_defaults import (
    PROPERTY_TEMPERATURE_SUPPLY,
    PROPERTY_TEMPERATURE_RETURN,
    PROPERTY_HEAT_DEMAND,
    PROPERTY_SET_PRESSURE,
)
from omotes_simulator_core.entities.assets.controller.controller_producer import ControllerProducer
from omotes_simulator_core.entities.assets.controller.controller_consumer import ControllerConsumer
from omotes_simulator_core.entities.assets.controller.controller_storage import ControllerStorage

logger = logging.getLogger(__name__)


class NetworkController(NetworkControllerAbstract):
    """Class to store the network controller."""

    def __init__(
        self,
        producers: List[ControllerProducer],
        consumers: List[ControllerConsumer],
        storages: List[ControllerStorage],
    ) -> None:
        """Constructor for controller for a heat network.

        The priority of the producers is set based on the marginal costs. The lowest marginal
        costs has the highest priority.

        :param List[ControllerProducer] producers: List of producers in the network.
        :param List[ControllerConsumer] consumers: List of consumers in the network.
        :param List[ControllerStorage] storages: List of storages in the network.
        """
        self.producers = producers
        self.consumers = consumers
        self.storages = storages
        self._set_priority_from_marginal_costs()

    def _set_priority_from_marginal_costs(self) -> None:
        """This method sets the priority of the producers based on the marginal costs.

        The priority of the producers is set based on the marginal costs. The producer with the
        lowest marginal costs has the highest priority (lowest value).
        """
        # Created a sorted list of unique marginal costs.
        unique_sorted_values = sorted(set([producer.marginal_costs for producer in self.producers]))

        # set the priority based on the index of the marginal cost in the sorted list.
        for producer in self.producers:
            producer.priority = unique_sorted_values.index(producer.marginal_costs) + 1

    def update_setpoints(self, time: datetime.datetime) -> dict:
        """Method to get the controller inputs for the network.

        :param float time: Time step for which to run the controller.
        :return: dict with the key the asset id and the heat demand for that asset.
        """
        # TODO add also the possibility to return mass flow rate instead of heat demand.
        if (
            self.get_total_supply() + (-1 * self.get_total_discharge_storage())
        ) <= self.get_total_demand(time):
            logger.warning(
                f"Total supply + storage is lower than total demand at time: {time}"
                f"Consumers are capped to the available power."
            )
            producers = self._set_producers_to_max()
            storages = self._set_all_storages_discharge_to_max()
            consumers = self._set_consumer_capped(time)
        else:
            # Consumers can meet their demand
            consumers = self._set_consumer_to_demand(time)
            if self.get_total_supply() > self.get_total_demand(time):
                surplus_supply = self.get_total_supply() - self.get_total_demand(time)
                if surplus_supply <= self.get_total_charge_storage():
                    storages = self._set_storages_power(surplus_supply)
                    producers = self._set_producers_to_max()
                else:
                    # need to cap the power of the source based on priority
                    storages = self._set_storages_power(self.get_total_charge_storage())
                    producers = self._set_producers_based_on_priority(time)

            else:
                deficit_supply = self.get_total_supply() - self.get_total_demand(time)
                storages = self._set_storages_power(deficit_supply)
                producers = self._set_producers_to_max()

        producers.update(consumers)
        producers.update(storages)
        return producers

    def get_total_demand(self, time: datetime.datetime) -> float:
        """Method to get the total heat demand of the network.

        :param datetime.datetime time: Time for which to get the total heat demand.
        :return float: Total heat demand of all consumers.
        """
        return sum([consumer.get_heat_demand(time) for consumer in self.consumers])

    def get_total_discharge_storage(self) -> float:
        """Method to get the total storage discharge power of the network.

        :return float: Total heat discharge of all storages.
        """
        # TODO add limit based on state of charge
        return float(sum([storage.max_discharge_power for storage in self.storages]))

    def get_total_charge_storage(self) -> float:
        """Method to get the total storage charge power of the network.

        :return float: Total heat charge of all storages.
        """
        # TODO add limit based on state of charge
        return float(sum([storage.max_charge_power for storage in self.storages]))

    def get_total_supply(self) -> float:
        """Method to get the total heat supply of the network.

        :return float: Total heat supply of all producers.
        """
        return float(sum([producer.power for producer in self.producers]))

    def get_total_supply_priority(self, priority: int) -> float:
        """Method to get the total supply of the network for all sources with a certain priority.

        :param int priority: The priority of the sources.
        :return float: Total heat supply of all producers with a certain priority.
        """
        return float(
            sum([producer.power for producer in self.producers if producer.priority == priority])
        )

    def _set_producers_to_max(self) -> dict:
        """Method to set the producers to the max power.

        :return dict: Dict with key= asset-id and value=setpoints for the producers.
        """
        producers = {}
        for source in self.producers:
            producers[source.id] = {
                PROPERTY_HEAT_DEMAND: source.power,
                PROPERTY_TEMPERATURE_RETURN: source.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: source.temperature_supply,
                PROPERTY_SET_PRESSURE: False,
            }
        # setting the first producer to set the pressure.
        producers[self.producers[0].id][PROPERTY_SET_PRESSURE] = True
        return producers

    def _get_basic_producers(self) -> dict:
        """Method to get the basic dict with setting for the producers.

        :return dict: Dict with key= asset-id and value=setpoints for the producers.
        """
        producers = {}
        for source in self.producers:
            producers[source.id] = {
                PROPERTY_HEAT_DEMAND: 0.0,
                PROPERTY_TEMPERATURE_RETURN: source.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: source.temperature_supply,
                PROPERTY_SET_PRESSURE: False,
            }
        return producers

    def _set_all_storages_discharge_to_max(self) -> dict:
        """Method to set all the storages to the max discharge power.

        :return dict: Dict with key= asset-id and value=setpoints for the storages.
        """
        storages = {}
        for storage in self.storages:
            storages[storage.id] = {
                PROPERTY_HEAT_DEMAND: storage.max_discharge_power,
                PROPERTY_TEMPERATURE_RETURN: storage.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: storage.temperature_supply,
            }
        return storages

    def _set_all_storages_charge_to_max(self) -> dict:
        """Method to set all the storages to the max charge power.

        :return dict: Dict with key= asset-id and value=setpoints for the storages.
        """
        storages = {}
        for storage in self.storages:
            storages[storage.id] = {
                PROPERTY_HEAT_DEMAND: storage.max_charge_power,
                PROPERTY_TEMPERATURE_RETURN: storage.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: storage.temperature_supply,
            }
        return storages

    def _set_storages_power(self, power: float = 0) -> dict:
        """Method to set the storages to power. discharge (-), charge (+).

        :return dict: Dict with key= asset-id and value=setpoints for the storages.
        """
        storages = {}
        for storage in self.storages:
            storages[storage.id] = {
                PROPERTY_HEAT_DEMAND: power / len(self.storages),
                PROPERTY_TEMPERATURE_RETURN: storage.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: storage.temperature_supply,
            }
        return storages

    def _set_consumer_capped(self, time: datetime.datetime) -> dict:
        """Method to set the consumer to the max available power of the producers.

        :param datetime.datetime time: Time for which to cap the heat demand based on available
        power.

        :return: dict with the key the asset id and the value a dict with the set points for the
        consumers.
        """
        factor = self.get_total_supply() / self.get_total_demand(time)
        return self._set_consumer_to_demand(time=time, factor=factor)

    def _set_consumer_to_demand(self, time: datetime.datetime, factor: float = 1.0) -> dict:
        """Method to set the consumer to the demand.

        :param datetime.datetime time: Time for which to set the consumer to the demand.
        :param float factor: Factor to multiply the heat demand with.

        :return: dict with the key the asset id and the value a dict with the set points for the
        consumers.
        """
        consumers = {}
        for consumer in self.consumers:
            consumers[consumer.id] = {
                PROPERTY_HEAT_DEMAND: consumer.get_heat_demand(time) * factor,
                PROPERTY_TEMPERATURE_RETURN: consumer.temperature_return,
                PROPERTY_TEMPERATURE_SUPPLY: consumer.temperature_supply,
            }
        return consumers

    def _set_producers_based_on_priority(self, time: datetime.datetime) -> dict:
        """Method to set the producers based on priority.

        :param datetime.datetime time: Time for which to set the producers based on priority.

        :return: dict with the key the asset id and the value a dict with the set points for the
        producers.
        """
        producers = self._get_basic_producers()
        total_demand = self.get_total_demand(time) + self.get_total_charge_storage()
        if total_demand > self.get_total_supply():
            logger.warning(
                "Total demand is higher than total supply. Cannot set producers based on priority."
            )

        priority = 0
        set_pressure = True
        while total_demand > 0:
            priority += 1
            total_supply = self.get_total_supply_priority(priority)
            priority_producers = [
                producer for producer in self.producers if producer.priority == priority
            ]
            if total_supply > total_demand:
                factor = total_demand / total_supply
                for source in priority_producers:
                    producers[source.id][PROPERTY_HEAT_DEMAND] = source.power * factor
                    if set_pressure:
                        producers[source.id][PROPERTY_SET_PRESSURE] = True
                        set_pressure = False
                total_demand = 0
            else:
                for source in priority_producers:
                    producers[source.id][PROPERTY_HEAT_DEMAND] = source.power
                total_demand -= total_supply
        if set_pressure:
            # set pressure has not been set and it needs to be set. This is set for the first
            # producer
            producers[self.producers[0].id][PROPERTY_SET_PRESSURE] = True
        return producers



#  Copyright (c) 2024. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing the classes for the controller."""

import datetime

import pandas as pd
import logging
from omotes_simulator_core.entities.assets.controller.asset_controller_abstract import (
    AssetControllerAbstract,
)

logger = logging.getLogger(__name__)


class ControllerConsumer(AssetControllerAbstract):
    """Class to store the consumer for the controller asset."""

    def __init__(
        self,
        name: str,
        identifier: str,
        temperature_supply: float,
        temperature_return: float,
        max_power: float,
        profile: pd.DataFrame,
    ):
        """Constructor for the consumer.

        :param str name: Name of the consumer.
        :param str identifier: Unique identifier of the consumer.
        """
        super().__init__(name, identifier)
        self.temperature_return = temperature_return
        self.temperature_supply = temperature_supply
        self.profile: pd.DataFrame = profile
        self.start_index = 0
        self.max_power: float = max_power

    def get_heat_demand(self, time: datetime.datetime) -> float:
        """Method to get the heat demand of the consumer.

        :param datetime.datetime time: Time for which to get the heat demand.
        :return: float with the heat demand.
        """
        for index in range(self.start_index, len(self.profile)):
            if abs((self.profile["date"][index].to_pydatetime() - time).total_seconds()) < 3600:
                self.start_index = index
                if self.profile["values"][index] > self.max_power:
                    logging.warning(
                        f"Demand of {self.name} is higher than maximum power of asset"
                        f" at time {time}."
                    )
                    return self.max_power
                else:
                    return float(self.profile["values"][index])
        return 0



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""atesCluster class."""
from typing import Dict
import os
import math

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.asset_defaults import (
    DEFAULT_TEMPERATURE,
    DEFAULT_TEMPERATURE_DIFFERENCE,
    PROPERTY_HEAT_DEMAND,
    PROPERTY_MASSFLOW,
    PROPERTY_PRESSURE_RETURN,
    PROPERTY_PRESSURE_SUPPLY,
    PROPERTY_TEMPERATURE_RETURN,
    PROPERTY_TEMPERATURE_SUPPLY,
)
from omotes_simulator_core.entities.assets.asset_defaults import ATES_DEFAULTS

from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.solver.network.assets.production_asset import ProductionAsset
from omotes_simulator_core.entities.assets.utils import (
    heat_demand_and_temperature_to_mass_flow,
)

path = os.path.dirname(__file__)
import jnius_config  # noqa

jnius_config.add_classpath(os.path.join(path, "bin/jfxrt.jar"))
jnius_config.add_classpath(os.path.join(path, "bin/rosim-batch-0.4.2.jar"))
from jnius import autoclass  # noqa

javaioFile = autoclass("java.io.File")
RosimSequential = autoclass("tno.calc.RosimSequential")


class AtesCluster(AssetAbstract):
    """A AtesCluster represents an asset that consumes heat and produces heat."""

    temperature_supply: float
    """The supply temperature of the asset [K]."""

    temperature_return: float
    """The return temperature of the asset [K]."""

    thermal_power_allocation: float
    """The thermal for injection (positive) or production (negative) by the asset [W]."""

    mass_flowrate: float
    """The flow rate going in or out by the asset [kg/s]."""

    def __init__(self, asset_name: str, asset_id: str, port_ids: list[str]):
        """Initialize a AtesCluster object.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        """
        super().__init__(asset_name=asset_name, asset_id=asset_id, connected_ports=port_ids)

        self.temperature_supply = DEFAULT_TEMPERATURE
        self.temperature_return = DEFAULT_TEMPERATURE - DEFAULT_TEMPERATURE_DIFFERENCE
        self.thermal_power_allocation = 0  # Watt
        self.mass_flowrate = 0  # kg/s
        self.solver_asset = ProductionAsset(name=self.name, _id=self.asset_id)
        # ATES default properties
        self.aquifer_depth = ATES_DEFAULTS.aquifer_depth  # meters
        self.aquifer_thickness = ATES_DEFAULTS.aquifer_thickness  # meters
        self.aquifer_mid_temperature = ATES_DEFAULTS.aquifer_mid_temperature  # Celcius
        self.aquifer_net_to_gross = ATES_DEFAULTS.aquifer_net_to_gross  # percentage
        self.aquifer_porosity = ATES_DEFAULTS.aquifer_porosity  # percentage
        self.aquifer_permeability = ATES_DEFAULTS.aquifer_permeability  # mD
        self.aquifer_anisotropy = ATES_DEFAULTS.aquifer_anisotropy  # -
        self.salinity = ATES_DEFAULTS.salinity  # ppm
        self.well_casing_size = ATES_DEFAULTS.well_casing_size  # inch
        self.well_distance = ATES_DEFAULTS.well_distance  # meters
        self.maximum_flow_charge = ATES_DEFAULTS.maximum_flow_charge  # m3/h
        self.maximum_flow_discharge = ATES_DEFAULTS.maximum_flow_discharge  # m3/h

        # Output list
        self.output: list = []

    def _calculate_massflowrate(self) -> None:
        """Calculate mass flowrate of the asset."""
        self.mass_flowrate = heat_demand_and_temperature_to_mass_flow(
            self.thermal_power_allocation, self.temperature_supply, self.temperature_return
        )

    def _set_solver_asset_setpoint(self) -> None:
        """Set the setpoint of solver asset."""
        if self.mass_flowrate > 0:
            self.solver_asset.supply_temperature = self.temperature_return
        else:
            self.solver_asset.supply_temperature = self.temperature_supply
        self.solver_asset.mass_flow_rate_set_point = self.mass_flowrate  # type: ignore

    def set_setpoints(self, setpoints: Dict) -> None:
        """Placeholder to set the setpoints of an asset prior to a simulation.

        :param Dict setpoints: The setpoints that should be set for the asset.
            The keys of the dictionary are the names of the setpoints and the values are the values
        """
        # Default keys required
        necessary_setpoints = {
            PROPERTY_TEMPERATURE_SUPPLY,
            PROPERTY_TEMPERATURE_RETURN,
            PROPERTY_HEAT_DEMAND,
        }
        # Dict to set
        setpoints_set = set(setpoints.keys())
        # Check if all setpoints are in the setpoints
        if necessary_setpoints.issubset(setpoints_set):
            self.thermal_power_allocation = setpoints[PROPERTY_HEAT_DEMAND]
            self.temperature_return = setpoints[PROPERTY_TEMPERATURE_RETURN]
            self.temperature_supply = setpoints[PROPERTY_TEMPERATURE_SUPPLY]

            self._calculate_massflowrate()
            self._run_rosim()
            self._set_solver_asset_setpoint()
        else:
            # Print missing setpoints
            raise ValueError(
                f"The setpoints {necessary_setpoints.difference(setpoints_set)} are missing."
            )

    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Method to add physical data to the asset.

        :param EsdlAssetObject esdl_asset: The esdl asset object to add the physical data from.
         :return:
        """
        self.aquifer_depth, _ = esdl_asset.get_property(
            esdl_property_name="aquiferTopDepth", default_value=self.aquifer_depth
        )
        self.aquifer_thickness, _ = esdl_asset.get_property(
            esdl_property_name="aquiferThickness", default_value=self.aquifer_thickness
        )
        self.aquifer_mid_temperature, _ = esdl_asset.get_property(
            esdl_property_name="aquiferMidTemperature", default_value=self.aquifer_mid_temperature
        )
        self.aquifer_net_to_gross, _ = esdl_asset.get_property(
            esdl_property_name="aquiferNetToGross", default_value=self.aquifer_net_to_gross
        )
        self.aquifer_porosity, _ = esdl_asset.get_property(
            esdl_property_name="aquiferPorosity", default_value=self.aquifer_porosity
        )
        self.aquifer_permeability, _ = esdl_asset.get_property(
            esdl_property_name="aquiferPermeability", default_value=self.aquifer_permeability
        )
        self.aquifer_anisotropy, _ = esdl_asset.get_property(
            esdl_property_name="aquiferAnisotropy", default_value=self.aquifer_anisotropy
        )
        self.salinity, _ = esdl_asset.get_property(
            esdl_property_name="salinity", default_value=self.salinity
        )
        self.well_casing_size, _ = esdl_asset.get_property(
            esdl_property_name="wellCasingSize", default_value=self.well_casing_size
        )
        self.well_distance, _ = esdl_asset.get_property(
            esdl_property_name="wellDistance", default_value=self.well_distance
        )

        maximum_charge_power, _ = esdl_asset.get_property(
            esdl_property_name="maxChargeRate", default_value=12e7
        )

        self.maximum_flow_charge = heat_demand_and_temperature_to_mass_flow(
            maximum_charge_power, self.temperature_supply, self.temperature_return
        )

        maximum_discharge_power, _ = esdl_asset.get_property(
            esdl_property_name="maxChargeRate", default_value=12e7
        )

        self.maximum_flow_discharge = heat_demand_and_temperature_to_mass_flow(
            maximum_discharge_power, self.temperature_supply, self.temperature_return
        )

        self._init_rosim()

    def write_to_output(self) -> None:
        """Placeholder to write the asset to the output.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        """
        output_dict = {
            PROPERTY_MASSFLOW: self.solver_asset.get_mass_flow_rate(1),
            PROPERTY_PRESSURE_SUPPLY: self.solver_asset.get_pressure(0),
            PROPERTY_PRESSURE_RETURN: self.solver_asset.get_pressure(1),
            PROPERTY_TEMPERATURE_SUPPLY: self.solver_asset.get_temperature(0),
            PROPERTY_TEMPERATURE_RETURN: self.solver_asset.get_temperature(1),
        }
        self.output.append(output_dict)

    def _init_rosim(self) -> None:
        """Function to initailized Rosim from XML file."""
        xmlfile = os.path.join(path, "bin/sequentialTemplate_v0.4.2_template.xml")
        with open(xmlfile, "r") as fd:
            xml_str = fd.read()

        # overwrite template value with ESDL properties for ROSIM input

        MODEL_TOP = self.aquifer_depth - 100
        AQUIFER_THICKNESS = self.aquifer_thickness
        NZ_AQUIFER = math.floor(AQUIFER_THICKNESS / 2)
        NZ = NZ_AQUIFER + 8
        AQUIFER_TOP = self.aquifer_depth
        AQUIFER_BASE = self.aquifer_depth + self.aquifer_thickness
        SURFACE_TEMPERATURE = self.aquifer_mid_temperature - 0.034 * (
            self.aquifer_depth + self.aquifer_thickness / 2
        )
        AQUIFER_NTG = self.aquifer_net_to_gross
        AQUIFER_PORO = self.aquifer_porosity
        AQUIFER_PERM_XY = self.aquifer_permeability
        AQUIFER_PERM_Z = AQUIFER_PERM_XY / self.aquifer_anisotropy
        SALINITY = self.salinity
        WELL2_X = self.well_distance + 300
        CASING_SIZE = self.well_casing_size

        xml_str = xml_str.replace("$NZ$", str(NZ))
        xml_str = xml_str.replace("$MODEL_TOP$", str(MODEL_TOP))
        xml_str = xml_str.replace("$TIME_STEP_UNIT$", str(2))
        xml_str = xml_str.replace("$WELL2_X$", str(WELL2_X))
        xml_str = xml_str.replace("$AQUIFER_TOP$", str(AQUIFER_TOP))
        xml_str = xml_str.replace("$AQUIFER_BASE$", str(AQUIFER_BASE))
        xml_str = xml_str.replace("$CASING_SIZE$", str(CASING_SIZE))
        xml_str = xml_str.replace("$SURFACE_TEMPERATURE$", str(SURFACE_TEMPERATURE))
        xml_str = xml_str.replace("$SALINITY$", str(SALINITY))
        xml_str = xml_str.replace("$NZ_AQUIFER$", str(NZ_AQUIFER))
        xml_str = xml_str.replace("$AQUIFER_THICKNESS$", str(AQUIFER_THICKNESS))
        xml_str = xml_str.replace("$AQUIFER_PORO$", str(AQUIFER_PORO))
        xml_str = xml_str.replace("$AQUIFER_NTG$", str(AQUIFER_NTG))
        xml_str = xml_str.replace("$AQUIFER_PERM_XY$", str(AQUIFER_PERM_XY))
        xml_str = xml_str.replace("$AQUIFER_PERM_Z$", str(AQUIFER_PERM_Z))

        temp_xmlfile_path = os.path.join(path, "bin/ates_sequential_temp.xml")
        with open(temp_xmlfile_path, "w") as temp_xmlfile:
            temp_xmlfile.write(xml_str)

        xmlfilejava = javaioFile(temp_xmlfile_path)
        self.rosim = RosimSequential(xmlfilejava, False, 2)

    def _run_rosim(self) -> None:
        """Function to calculate storage temperature after injection and production."""
        volume_flow = self.mass_flowrate * 3600 / 1027  # convert to second and hardcoded saline
        # density needs to change with PVT calculation
        timestep = 1  # HARDCODED to 1 hour

        rosim_input__flow = [volume_flow, -1 * volume_flow]  # first elemnt is for producer well
        # and second element is for injection well, positive flow is going upward and negative flow
        # is downward

        if volume_flow > 0:
            rosim_input_temperature = [self.temperature_supply - 273.15, -1]  # Celcius, -1 in
            # injection well to make sure it is not used
        elif volume_flow < 0:
            rosim_input_temperature = [-1, self.temperature_return - 273.15]  # Celcius, -1 in
            # producer well to make sure it is not used
        else:
            rosim_input_temperature = [-1, -1]  # -1 in both producer and injection well to make
            # sure it is not used

        ates_temperature = self.rosim.getTempsNextTimeStep(
            rosim_input__flow, rosim_input_temperature, timestep
        )

        hot_well_temperature = ates_temperature[0] + 273.15  # convert to K
        cold_well_temperature = ates_temperature[1] + 273.15  # convert to K

        # update supply return temperature from ATES
        self.temperature_supply = hot_well_temperature
        self.temperature_return = cold_well_temperature



#  Copyright (c) 2024. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains the HeatPump class."""

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.solver.network.assets.base_asset import BaseAsset


class HeatPump(AssetAbstract):
    """A HeatPump represents an asset that produces heat."""

    def __init__(self, asset_name: str, asset_id: str, port_ids: list[str]):
        """Initialize a HeatPump object.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        :param List[str] port_ids: List of ids of the connected ports.
        """
        super().__init__(asset_name=asset_name, asset_id=asset_id, connected_ports=port_ids)
        self.number_of_con_points = 4
        self.solver_asset = BaseAsset(
            name=self.name,
            _id=self.asset_id,
            number_of_unknowns=12,
            number_connection_points=self.number_of_con_points,
        )

    def set_setpoints(self, setpoints: dict) -> None:
        """Placeholder to set the setpoints of an asset prior to a simulation."""
        pass

    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Add physical data to the asset."""
        pass

    def write_to_output(self) -> None:
        """Write the output of the asset to the output list."""
        pass



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing the Fall type class."""
import numpy as np

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.base_asset import BaseAsset
from omotes_simulator_core.solver.matrix.index_core_quantity import index_core_quantity


class FallType(BaseAsset):
    """A class to represent a fall type asset in a network.

    This class inherits from the BaseAsset class and implements the methods to generate the
    equations for the fall type asset.

    Attributes
    ----------
    connected_nodes : list[int]
        A list of integers that store the node indices that are connected to the asset.
    number_of_connection_point : int
        The number of connection points for the asset.

    Methods
    -------
    get_equations() -> list[EquationObject]
        Returns a list of EquationObjects that represent the equations for the asset.
    add_equations()
        Adds the equations for the asset to the equation system.
    get_internal_cont_equation() -> EquationObject
        Returns an EquationObject that represents the internal continuity equation for the asset.
    get_internal_pressure_loss_equation() -> EquationObject
        Returns an EquationObject that represents the internal pressure loss equation for the asset.
    get_internal_energy_equation() -> EquationObject
        Returns an EquationObject that represents the internal energy equation for the asset.
    update_pressure_loss_equation() -> EquationObject
        Returns an updated EquationObject that represents the internal pressure loss equation
        for the asset.
    """

    def __init__(
        self,
        name: str,
        _id: str,
        supply_temperature: float = 293.15,
        heat_supplied: float = 0.0,
        loss_coefficient: float = 1.0,
    ):
        """
        Initializes the FallType object with the given parameters.

        Parameters
        ----------
        name : str The name of the asset.
        _id : str The unique identifier of the asset.
        number_of_unknowns : int, optional
            The number of unknown variables for the asset. The default is 6, which corresponds
            to the mass flow rate, pressure, and temperature at each connection point.
        number_con_points : int, optional
            The number of connection points for the asset. The default is 2, which corresponds to
            the inlet and outlet.
        """
        super().__init__(
            name=name,
            _id=_id,
            number_of_unknowns=index_core_quantity.number_core_quantities * 2,
            number_connection_points=2,
            supply_temperature=supply_temperature,
        )
        self.heat_supplied = heat_supplied
        self.loss_coefficient = loss_coefficient

    def get_equations(self) -> list[EquationObject]:
        """Returns a list of EquationObjects that represent the equations for the asset.

         The equations are:

        - Pressure balance at each connection point
        - Thermal balance at each connection point
        - Internal continuity equation
        - Internal pressure loss equation

        :return: list[EquationObject]
            A list of EquationObjects that contain the indices, coefficients, and right-hand side
            values of the equations.
        """
        # Check if there are two nodes connected to the asset
        if len(self.connected_nodes) != 2:
            raise ValueError("The number of connected nodes must be 2!")
        # Check if the number of unknowns is 6
        if self.number_of_unknowns != 6:
            raise ValueError("The number of unknowns must be 6!")
        # Get the equations for the asset
        equations = [
            self.get_press_to_node_equation(0),
            self.get_press_to_node_equation(1),
            self.get_thermal_equations(0),
            self.get_thermal_equations(1),
            self.get_internal_cont_equation(),
            self.get_internal_pressure_loss_equation(),
        ]
        return equations

    def get_thermal_equations(self, connection_point: int) -> EquationObject:
        """Gets a thermal equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the thermal equation.
        :rtype: EquationObject
        """
        if (
            self.prev_sol[
                self.get_index_matrix(
                    "mass_flow_rate", connection_point=connection_point, use_relative_indexing=True
                )
            ]
            > 1e-5
        ):
            return self.get_internal_energy_equation()
        else:
            return self.get_internal_energy_to_node_equation(connection_point=connection_point)

    def get_internal_cont_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the internal continuity equation for the asset.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side value
            of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    "mass_flow_rate", connection_point=0, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    "mass_flow_rate", connection_point=1, use_relative_indexing=False
                ),
            ]
        )
        equation_object.coefficients = np.array([1.0, 1.0])
        equation_object.rhs = 0.0
        return equation_object

    def get_internal_energy_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the internal energy equation for the asset.

         The equation is:

        - Mass flow rate at inlet * Specific internal energy at inlet +
            Mass flow rate at outlet * Specific internal energy at outlet - Heat supplied = 0

        :return:EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side value of
            the equation.
        """
        equation_object = EquationObject()
        self.update_heat_supplied()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=0, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    property_name="internal_energy", connection_point=0, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=1, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    property_name="internal_energy", connection_point=1, use_relative_indexing=False
                ),
            ]
        )
        equation_object.coefficients = np.array(
            [
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ],
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="mass_flow_rate",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ],
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=1,
                        use_relative_indexing=True,
                    )
                ],
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="mass_flow_rate",
                        connection_point=1,
                        use_relative_indexing=True,
                    )
                ],
            ]
        )
        equation_object.rhs = (
            self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=0, use_relative_indexing=True
                )
            ]
            * self.prev_sol[
                self.get_index_matrix(
                    property_name="internal_energy", connection_point=0, use_relative_indexing=True
                )
            ]
            + self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=1, use_relative_indexing=True
                )
            ]
            * self.prev_sol[
                self.get_index_matrix(
                    property_name="internal_energy", connection_point=1, use_relative_indexing=True
                )
            ]
            + self.heat_supplied
        )
        return equation_object

    def get_internal_pressure_loss_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the pressure loss equation for the asset.

        The equation is:
        - Pressure at inlet - Pressure at outlet - 2 * Loss coefficient * Mass flow rate *
        abs(Mass flow rate) = 0
        :return: EquationObject
        An EquationObject that contains the indices, coefficients, and right-hand side value
        of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=0, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    property_name="pressure", connection_point=0, use_relative_indexing=False
                ),
                self.get_index_matrix(
                    property_name="pressure", connection_point=1, use_relative_indexing=False
                ),
            ]
        )
        self.update_loss_coefficient()
        mass_flow_rate = self.prev_sol[
            self.get_index_matrix(
                property_name="mass_flow_rate", connection_point=0, use_relative_indexing=True
            )
        ]
        if mass_flow_rate < 1e-5:
            equation_object.coefficients = np.array(
                [-2.0 * self.loss_coefficient * 1e-5, -1.0, 1.0]
            )
            equation_object.rhs = -self.loss_coefficient * mass_flow_rate * 1e-5
        else:
            equation_object.coefficients = np.array(
                [-2.0 * self.loss_coefficient * abs(mass_flow_rate), -1.0, 1.0]
            )
            equation_object.rhs = -self.loss_coefficient * mass_flow_rate * abs(mass_flow_rate)
        return equation_object

    def update_loss_coefficient(self) -> None:
        """Basic function which does not do anything, but can be overwritten in derived classes."""

    def update_heat_supplied(self) -> None:
        """Basic function which does not do anything, but can be overwritten in derived classes."""



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing abstract BaseItem class."""
from abc import ABC, abstractmethod
from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.matrix.index_core_quantity import index_core_quantity
import numpy as np


class BaseItem(ABC):
    """A base class for items in a network."""

    prev_sol: np.ndarray
    """The previous solution of the calculation, defaults to an array of
    zeros with a certain length (number of unknowns)."""

    def __init__(self, number_of_unknowns: int, name: str, _id: str, number_connection_points: int):
        """Initializes the BaseItem object with the given parameters.

        :param int number_of_unknowns: The number of unknown variables for the item.
        :param str name: The name of the item.
        :param str _id: The unique identifier of the item.
        :param int number_connection_points: The number of connection points of the item.
        """
        self.name = name
        self.id = _id
        self.number_of_unknowns = number_of_unknowns
        self.number_of_connection_point = number_connection_points
        self.matrix_index = 0
        self.prev_sol = np.zeros(self.number_of_unknowns)

    def __repr__(self) -> str:
        """Returns the string representation of the item."""
        return str(self.name)

    def set_matrix_index(self, index: int) -> None:
        """Sets the matrix index of the item.

        :param int index: The index of the item in the matrix.
        """
        self.matrix_index = index

    def get_index_matrix(
        self, property_name: str, connection_point: int, use_relative_indexing: bool
    ) -> int:
        """Returns the index of the property&connection point in the coefficient array or matrix.

        :param str property_name: The property name for which the matrix index is needed.
        :param int connection_point: The connection point for which the matrix index is needed.
        :param bool use_relative_indexing: returns the index into coefficient array for the current
        componentthe matrix if true, otherwise return the index in the complete matrix if false.
        :return: The matrix index for the property and connection point.
        """
        return index_core_quantity.get_index_property(
            property_name=property_name, connection_point=connection_point
        ) + (0 if use_relative_indexing else self.matrix_index)

    @abstractmethod
    def get_equations(self) -> list[EquationObject]:
        """Returns the equations for the item.

        :return: The equations for the item.
        :rtype: list[EquationObject]
        """

    @abstractmethod
    def disconnect_node(self, connection_point: int) -> None:
        """Disconnects a node from the item.

        :param int connection_point: The connection point to disconnect.
        """



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing the abstract class for the mappers."""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar


Entity = TypeVar("Entity")
EsdlAssetObject = TypeVar("EsdlAssetObject")


class EsdlMapperAbstract(ABC, Generic[Entity, EsdlAssetObject]):
    """Abstract class to be used for deriving mapper classes from esdl to our own classes."""

    @abstractmethod
    def to_esdl(self, entity: Entity) -> EsdlAssetObject:
        """Map an Entity to a EsdlAsset."""

    @abstractmethod
    def to_entity(self, model: EsdlAssetObject) -> Entity:
        """Map an esdl asset  to an Entity."""



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing the network class."""
import uuid
import numpy.typing as npt

from omotes_simulator_core.solver.network.assets.base_asset import BaseAsset
from omotes_simulator_core.solver.network.assets.boundary import BaseBoundary
from omotes_simulator_core.solver.network.assets.fall_type import FallType
from omotes_simulator_core.solver.network.assets.node import Node
from omotes_simulator_core.solver.network.assets.production_asset import ProductionAsset
from omotes_simulator_core.solver.network.assets.solver_pipe import SolverPipe


class Network:
    """Class to store a network consisting of nodes and assets."""

    str_to_class_dict = {
        "Boundary": BaseBoundary,
        "Asset": BaseAsset,
        "Fall": FallType,
        "Production": ProductionAsset,
        "Pipe": SolverPipe,
    }
    assets: dict[str, BaseAsset]
    nodes: dict[str, Node]

    def __init__(self) -> None:
        """Constructor of the network class.

        Initializes the class properties and loads the fluid properties.
        """
        self.assets = {}
        self.nodes = {}

    def add_asset(
        self, asset_type: str, name: str | None = None, identifier: str | None = None
    ) -> str:
        """Method to add an asset to the network.

        This method creates and asset of the given type.
        If the type does not exist a ValueError is raised.
        The unique id which is created for this asset is returned.
        :param name: Unique name of the asset, if not given a random uuid is created.
        :param str asset_type: The type of asset to be added
        :param str identifier: Unique identifier of the asset, if not given a random uuid is
        created.
        :return: Unique id of the asset.
        """
        if asset_type not in self.str_to_class_dict:
            raise ValueError(asset_type + " not recognized.")
        if name is None:
            name = str(uuid.uuid4())
        if identifier is None:
            identifier = str(uuid.uuid4())
        self.assets[name] = self.str_to_class_dict[asset_type](name=name, _id=identifier)
        return name

    def add_existing_asset(self, asset: BaseAsset) -> str:
        """Method to add an existing asset to the network.

        This method adds an existing asset to the network. It checks if the asset already exists.
        If it does a ValueError is raised.
        :param BaseAsset asset: The asset to be added to the network.
        :return: Unique id of the asset.
        """
        if asset.name in self.assets:
            raise ValueError(f"{asset.name} already exists in network.")
        self.assets[asset.name] = asset
        return asset.name

    def _connect_single_asset_at_node(
        self,
        asset_id_connected: str,
        connection_point_connected: int,
        asset_id_unconnected: str,
        connection_point_unconnected: int,
    ) -> str:
        """Method to connect a single asset to a node.

        This method connects the unconnected asset to the node of the connected asset.
        The id of the node is returned.

        :param asset_id_connected: id of the connected asset
        :type asset_id_connected: str
        :param connection_point_connected: Connection point of the connected asset
        :type connection_point_connected: int
        :param asset_id_unconnected: id of the unconnected asset
        :type asset_id_unconnected: str
        :param connection_point_unconnected: Connection point of the unconnected asset
        :type connection_point_unconnected: int
        :return: id of node connecting the two assets
        """
        # Get the node of the connected asset
        node = self.assets[asset_id_connected].get_connected_node(
            connection_point=connection_point_connected
        )
        # Connect the asset to the node
        self.assets[asset_id_unconnected].connect_node(
            connection_point=connection_point_unconnected, node=node
        )
        # Connect the node to the asset
        node.connect_asset(
            self.assets[asset_id_unconnected], connection_point=connection_point_unconnected
        )
        return node.name

    def _connect_both_assets_at_node(
        self,
        asset1_id: str,
        connection_point_1: int,
        asset2_id: str,
        connection_point_2: int,
    ) -> str:
        """Method to connect to assets at the given connection points.

        Connects the two assets to a new node and returns the id of the node.

        :param asset1_id: id of first asset to be connected
        :type asset1_id: str
        :param connection_point_1: Connection point of first asset to be connected
        :type connection_point_1: int
        :param asset2_id: id of second asset to be connected
        :type asset2_id: str
        :param connection_point_2: Connection point of second asset to be connected
        :type connection_point_2: int
        :return: id of node connecting the two assets
        """
        # Create a new node
        node_id = str(uuid.uuid4())
        self.nodes[node_id] = Node(name=node_id, _id=node_id)
        # Connect the assets to the node
        for asset_id, connection_point in [
            (asset1_id, connection_point_1),
            (asset2_id, connection_point_2),
        ]:
            self.assets[asset_id].connect_node(
                connection_point=connection_point, node=self.nodes[node_id]
            )
            self.nodes[node_id].connect_asset(
                asset=self.assets[asset_id], connection_point=connection_point
            )
        return node_id

    def _connect_both_assets_and_replace_node(
        self,
        asset1_id: str,
        connection_point_1: int,
        asset2_id: str,
        connection_point_2: int,
    ) -> str:
        """Method to connect to assets at the given connection points.

        Connects the two assets to a new node and returns the id of the node.

        :param asset1_id: id of first asset to be connected
        :type asset1_id: str
        :param connection_point_1: Connection point of first asset to be connected
        :type connection_point_1: int
        :param asset2_id: id of second asset to be connected
        :type asset2_id: str
        :param connection_point_2: Connection point of second asset to be connected
        :type connection_point_2: int
        :return: id of node connecting the two assets
        """
        # Retrieve the nodes of the assets
        node1 = self.assets[asset1_id].get_connected_node(connection_point=connection_point_1)
        node2 = self.assets[asset2_id].get_connected_node(connection_point=connection_point_2)
        # Check if the nodes are the same, if so return the name of the node.
        if node1 == node2:
            return node1.name
        else:
            # Both nodes are different, connect the two nodes and remove the second node.
            # First connect all assets connected to the second node to the first node.
            for connected_comp, connection_point in node2.get_connected_assets():
                # Disconnect the connected component from the second node
                connected_comp.disconnect_node(connection_point=connection_point)
                # Connect the connected component to the first node
                self.connect_assets(
                    asset1_id=connected_comp.name,
                    connection_point_1=connection_point,
                    asset2_id=asset1_id,
                    connection_point_2=connection_point_1,
                )
            # Remove the second node from the network
            del self.nodes[node2.name]
            # Finally connect the first asset to the second asset
            self.assets[asset2_id].disconnect_node(connection_point=connection_point_2)
            self.assets[asset2_id].connect_node(connection_point=connection_point_2, node=node1)
            return node1.name

    def connect_assets(
        self,
        asset1_id: str,
        connection_point_1: int,
        asset2_id: str,
        connection_point_2: int,
    ) -> str:
        """Method to connect to assets at the given connection points.

        This method connects the two assets if the exists. It checks if they already are connected.
        If not a new node is created. Otherwise, both are connected to the existing node.
        The id of the node connecting the two is returned

        :param asset1_id: id of first asset to be connected
        :param connection_point_1: Connection point of first asset to be connected
        :param asset2_id: id of second asset to be connected
        :param connection_point_2: Connection point of second asset to be connected
        :return: id of node connecting the two assets
        """
        # Check if both assets exist; if not raise a ValueError
        for asset_id in [asset1_id, asset2_id]:
            self.exists_asset(asset_id=asset_id)
        # Create boolean list with for each asset wether the connection point is connected
        connected_1 = self.assets[asset1_id].is_connected(connection_point=connection_point_1)
        connected_2 = self.assets[asset2_id].is_connected(connection_point=connection_point_2)
        # Use the connect_assets_decision_tree to connect the assets
        return self.connect_assets_decision_tree(
            connected_1=connected_1,
            connected_2=connected_2,
            asset1_id=asset1_id,
            connection_point_1=connection_point_1,
            asset2_id=asset2_id,
            connection_point_2=connection_point_2,
        )

    def connect_assets_decision_tree(
        self,
        connected_1: bool,
        connected_2: bool,
        asset1_id: str,
        connection_point_1: int,
        asset2_id: str,
        connection_point_2: int,
    ) -> str:
        """Method to connect to assets at the given connection points.

        The method uses a decision tree to connect the assets. The decision tree is as follows:
        1. If both assets are not connected, create a new node and connect everything.
        2. If asset 1 is connected and asset 2 is not, connect the node of asset 1 to asset 2.
        3. If asset 2 is connected and asset 1 is not, connect the node of asset 2 to asset 1.
        4. If both assets are connected, check if they are connected to the same node, otherwise
        connect the two nodes and remove the second node.
        5. If none of the above is true raise a NotImplementedError.

        :param connected_1: Boolean indicating if asset 1 is connected
        :param connected_2: Boolean indicating if asset 2 is connected
        :param asset1_id: id of first asset to be connected
        :param connection_point_1: Connection point of first asset to be connected
        :param asset2_id: id of second asset to be connected
        :param connection_point_2: Connection point of second asset to be connected
        :return: id of node connecting the two assets
        """
        if not any([connected_1, connected_2]):
            # both assets are not connected. Create a new node and connect everything.
            return self._connect_both_assets_at_node(
                asset1_id=asset1_id,
                connection_point_1=connection_point_1,
                asset2_id=asset2_id,
                connection_point_2=connection_point_2,
            )
        elif connected_1 and (not connected_2):
            # asset 1 connected asset 2 not, connect the node of asset 1 to asset 2
            return self._connect_single_asset_at_node(
                asset_id_connected=asset1_id,
                connection_point_connected=connection_point_1,
                asset_id_unconnected=asset2_id,
                connection_point_unconnected=connection_point_2,
            )
        elif (not connected_1) and connected_2:
            # asset 2 connected asset 1 not connect the node of asset 2 to asset 1
            return self._connect_single_asset_at_node(
                asset_id_connected=asset2_id,
                connection_point_connected=connection_point_2,
                asset_id_unconnected=asset1_id,
                connection_point_unconnected=connection_point_1,
            )
        else:
            # Effectively we call: all([connected_1, connected_2])
            # both assets are connected, check if they are connected to the same node
            # otherwise connect the two nodes and remove the second node.
            return self._connect_both_assets_and_replace_node(
                asset1_id=asset1_id,
                connection_point_1=connection_point_1,
                asset2_id=asset2_id,
                connection_point_2=connection_point_2,
            )

    def exists_asset(self, asset_id: str) -> bool:
        """Method returns true when an asset with the given id exists in the network.

        :param str asset_id: unique id of the asset to check.

        :return:True when asset exists and False when not
        """
        if not (asset_id in self.assets):
            raise ValueError(f"Asset with id:{asset_id} does not exist in network.")
        else:
            return True

    def exists_node(self, node_id: str) -> bool:
        """Method returns true when a node with the given id exists in the network.

        :param str node_id: unique id of the node to check.

        :return:True when node exists and False when not
        """
        if not (node_id in self.nodes):
            raise ValueError(f"Node with id:{node_id} does not exist in network.")
        else:
            return True

    def remove_asset(self) -> None:
        """Method to remove an asset from the network."""

    def disconnect_asset(self) -> None:
        """Method to disconnect an asset from the network."""

    def get_asset(self, asset_id: str) -> BaseAsset:  # type: ignore
        """Method to get an asset in the network.

        Method returns the asset with the given id, when it exists in the network.
        when it does not exist a ValueError is raised.

        :param id: asset_id of the asset which needs to be retrieved.
        :type asset_id: str
        :return: Asset
        """
        if self.exists_asset(asset_id):
            return self.assets[asset_id]

    def get_node(self, node_id: str) -> Node:  # type: ignore
        """Method to get a node in the network.

        Method returns the node with the given id, when it exists in the network.
        when it does not exist a ValueError is raised.

        :param node_id: node_id of the node which needs to be retrieved.
        :type node_id: str
        :return: Node
        """
        if self.exists_node(node_id=node_id):
            return self.nodes[node_id]

    def check_connectivity_assets(self) -> bool:
        """Method to check if all assets are connected at all of their connection points.

        Method returns True when all assets are connected and False when an asset is not connected.
        :return: True or False depending on if all assets are connected
        """
        # TODO pass back which assets are not connected.
        result = [asset.is_all_connected() for _, asset in self.assets.items()]
        return all(result)

    def check_connectivity_nodes(self) -> bool:
        """Method to check if all nodes are connected.

        Method returns True when all nodes are connected and False when an node is not connected.
        :return: True or False depending on if all nodes are connected
        """
        result = [node.is_connected() for _, node in self.nodes.items()]
        return all(result)

    def check_connectivity(self) -> bool:
        """Method to check if all nodes and assets are connected.

        :return:True when everything is connected and False when not.
        """
        return self.check_connectivity_assets() and self.check_connectivity_nodes()

    def set_result_asset(self, solution: npt.NDArray) -> None:
        """Method to transfer the solution to the asset in the network.

        :param list[float] solution:Solution to be transferred to the assets.
        :return: None
        """
        for asset in self.assets:
            index = self.get_asset(asset_id=asset).matrix_index
            nou = self.get_asset(asset_id=asset).number_of_unknowns
            self.get_asset(asset_id=asset).prev_sol = solution[index : index + nou].tolist()

    def set_result_node(self, solution: npt.NDArray) -> None:
        """Method to transfer the solution to the nodes in the network.

        :param list[float] solution:Solution to be transferred to the nodes.
        :return: None
        """
        for node in self.nodes:
            index = self.get_node(node_id=node).matrix_index
            nou = self.get_node(node_id=node).number_of_unknowns
            self.get_node(node_id=node).prev_sol = solution[index : index + nou].tolist()

    def print_result(self) -> None:
        """Method to print the result of the network."""
        for asset in self.assets:
            print(type(self.get_asset(asset_id=asset)))
            print(self.get_asset(asset_id=asset).get_result())



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract class for asset."""

from abc import ABC, abstractmethod
from typing import Dict, List

from pandas import DataFrame, concat

from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.solver.network.assets.base_asset import BaseAsset

from omotes_simulator_core.entities.assets.asset_defaults import (
    PROPERTY_MASSFLOW,
    PROPERTY_PRESSURE,
    PROPERTY_TEMPERATURE,
)


class AssetAbstract(ABC):
    """Abstract class for Asset."""

    name: str
    """The name of the asset."""

    asset_id: str
    """The unique identifier of the asset."""

    outputs: List[List[Dict[str, float]]]
    """The output of the asset as a list with a dictionary per timestep."""

    connected_ports: List[str]
    """List of ids of the connected ports."""
    solver_asset: BaseAsset
    """The asset object use for the solver."""
    asset_type = "asset_abstract"
    """The type of the asset."""
    number_of_con_points: int = 2
    """The number of connection points of the asset."""

    def __init__(self, asset_name: str, asset_id: str, connected_ports: List[str]) -> None:
        """Basic constructor for asset objects.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        :param List[str] connected_ports: List of ids of the connected ports.
        """
        self.from_junction = None
        self.to_junction = None
        self.name = asset_name
        self.asset_id = asset_id
        self.connected_ports = connected_ports
        self.outputs = [[] for _ in range(len(self.connected_ports))]

    def __repr__(self) -> str:
        """Method to print string with the name of the asset."""
        return self.__class__.__name__ + " " + self.name

    @abstractmethod
    def set_setpoints(self, setpoints: Dict) -> None:
        """Placeholder to set the setpoints of an asset prior to a simulation.

        :param Dict setpoints: The setpoints that should be set for the asset.
            The keys of the dictionary are the names of the setpoints and the values are the values
        """

    def get_setpoints(self) -> Dict[str, float]:
        """Placeholder to get the setpoint attributes of an asset.

        :return Dict: The setpoints of the asset. The keys of the dictionary are the names of the
            setpoints and the values are the values.
        """
        return {}

    @abstractmethod
    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Placeholder method to add physical data to an asset."""

    def write_standard_output(self) -> None:
        """Write the output of the asset to the output list.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        The output of the asset is a list with a dictionary for each port
        of the asset. Teh basic properties mass flow rate, pressure and temperature are stored.
        All assets can add their own properties to the dictionary.
        """
        for i in range(len(self.connected_ports)):
            output_dict_temp = {
                PROPERTY_MASSFLOW: self.solver_asset.get_mass_flow_rate(i),
                PROPERTY_PRESSURE: self.solver_asset.get_pressure(i),
                PROPERTY_TEMPERATURE: self.solver_asset.get_temperature(i),
            }
            self.outputs[i].append(output_dict_temp)

    @abstractmethod
    def write_to_output(self) -> None:
        """Placeholder to get data and store it in the asset."""

    def get_timeseries(self) -> DataFrame:
        """Get timeseries as a dataframe from a asset.

        The header is a tuple of the port id and the property name.
        """
        # Create dataframe

        temp_data = DataFrame()
        for i in range(len(self.connected_ports)):
            temp_frame = DataFrame(self.outputs[i])
            temp_frame.columns = [
                (self.connected_ports[i], column_name) for column_name in temp_frame.columns
            ]
            temp_data = concat([temp_data, temp_frame], axis=1)
        return temp_data



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing utility functions."""

from esdl.esdl_handler import EnergySystemHandler
from pathlib import Path


def pyesdl_from_file(file_path: str | Path) -> EnergySystemHandler:
    """
    Loads esdl file into memory and returns a handle to be able to use this file.

    Please note that it is not checked if the file is a valid esdl file.
    :param str | Path file_path: Location of the ESDL file to be loaded into memory.
    If file_path is a string, convert it to a Path object, then check if its not a path.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not file_path.is_file():
        raise FileNotFoundError("Not a valid path to an ESDL file")
    esh = EnergySystemHandler()
    esh.load_file(str(file_path))
    return esh


def pyesdl_from_string(input_str: str) -> EnergySystemHandler:
    """
    Loads esdl file from a string into memory.

    Please note that it is not checked if the contents of the string is a valid esdl.
    :param input_str: string containing the contents of an esdl file.
    """
    esh = EnergySystemHandler()
    esh.load_from_string(input_str)
    return esh
